package files

import (
	"archive/tar"
	"compress/gzip"
	"io"
	"os"
	"path"
	"path/filepath"
)

// untargz a tarball to a target, from
// http://blog.ralch.com/tutorial/golang-working-with-tar-and-gzipf
func UnTargz(tarball, target string, onlyFiles []string) error {
	zreader, err := os.Open(tarball)
	if err != nil {
		return err
	}
	defer zreader.Close() //nolint:errcheck

	reader, err := gzip.NewReader(zreader)
	defer reader.Close() //nolint:errcheck
	if err != nil {
		return err
	}

	tarReader := tar.NewReader(reader)

	for {
		inkey := false
		header, err := tarReader.Next()
		if err == io.EOF {
			break
		} else if err != nil {
			return err
		}

		for _, value := range onlyFiles {
			if value == "*" || value == path.Base(header.Name) {
				inkey = true
				break
			}
		}

		if !inkey && len(onlyFiles) > 0 {
			continue
		}

		p := filepath.Join(target, path.Base(header.Name))
		err = UnTarFile(header, p, tarReader)
		if err != nil {
			return err
		}
	}
	return nil
}

// untargz a tarball to a target including any folders inside the tarball
// http://blog.ralch.com/tutorial/golang-working-with-tar-and-gzipf
func UnTargzAll(tarball, target string) error {
	zreader, err := os.Open(tarball)
	if err != nil {
		return err
	}
	defer zreader.Close()

	reader, err := gzip.NewReader(zreader)
	defer reader.Close()
	if err != nil {
		panic(err)
	}

	tarReader := tar.NewReader(reader)

	for {
		header, err := tarReader.Next()
		if err == io.EOF {
			break
		} else if err != nil {
			return err
		}

		p := filepath.Join(target, header.Name)
		err = UnTarFile(header, p, tarReader)
		if err != nil {
			return err
		}
	}
	return nil
}

// UnTarFile extracts one file from the tar, or creates a directory
func UnTarFile(header *tar.Header, target string, tarReader io.Reader) error {
	info := header.FileInfo()
	if info.IsDir() {
		if err := os.MkdirAll(target, info.Mode()); err != nil {
			return err
		}
		return nil
	}
	// In a normal archive, directories are mentionned before their files
	// But in an archive generated by helm, no directories are mentionned
	if err := os.MkdirAll(path.Dir(target), 0755); err != nil {
		return err
	}

	file, err := os.OpenFile(target, os.O_CREATE|os.O_TRUNC|os.O_WRONLY, info.Mode())
	if err != nil {
		return err
	}
	defer file.Close()
	_, err = io.Copy(file, tarReader)
	return err
}
